package jdk.net;

import java.security.AccessController;
import java.security.PrivilegedAction;
import sun.misc.SharedSecrets;
import java.net.SocketException;
import java.io.FileDescriptor;
import java.util.Collections;
import java.util.HashSet;
import sun.misc.JavaIOFileDescriptorAccess;
import java.util.Set;
import java.net.SocketOption;
import jdk.Exported;

@Exported
public final class ExtendedSocketOptions
{
    public static final SocketOption<SocketFlow> SO_FLOW_SLA;
    public static final SocketOption<Integer> TCP_KEEPIDLE;
    public static final SocketOption<Integer> TCP_KEEPINTERVAL;
    public static final SocketOption<Integer> TCP_KEEPCOUNT;
    private static final ExtendedSocketOptions.PlatformSocketOptions platformSocketOptions;
    private static final boolean flowSupported;
    private static final boolean keepAliveOptSupported;
    private static final Set<SocketOption<?>> extendedOptions;
    private static final JavaIOFileDescriptorAccess fdAccess;
    
    private ExtendedSocketOptions() {
    }
    
    static Set<SocketOption<?>> options() {
        HashSet<ExtendedSocketOptions> s = new HashSet();
        if (ExtendedSocketOptions.flowSupported) {
            s.add(ExtendedSocketOptions.SO_FLOW_SLA);
        }
        if (ExtendedSocketOptions.keepAliveOptSupported) {
            s.add(ExtendedSocketOptions.TCP_KEEPCOUNT);
            s.add(ExtendedSocketOptions.TCP_KEEPIDLE);
            s.add(ExtendedSocketOptions.TCP_KEEPINTERVAL);
        }
        return Collections.unmodifiableSet(s);
    }
    
    private static <T> T checkValueType(Object o, Class<T> obj) {
        if (!obj.isAssignableFrom(o.getClass())) {
            throw new IllegalArgumentException("Found: " + o.getClass() + ", Expected: " + obj);
        }
        return o;
    }
    
    public static void setFlowOption(FileDescriptor fileDescriptor, SocketFlow socketFlow) throws SocketException {
        socketFlow.status(ExtendedSocketOptions.PlatformSocketOptions.get().setFlowOption(ExtendedSocketOptions.fdAccess.get(fileDescriptor), socketFlow.priority(), socketFlow.bandwidth()));
    }
    
    public static int getFlowOption(FileDescriptor fileDescriptor, SocketFlow socketFlow) throws SocketException {
        return ExtendedSocketOptions.PlatformSocketOptions.get().getFlowOption(ExtendedSocketOptions.fdAccess.get(fileDescriptor), socketFlow);
    }
    
    private static boolean flowSupported() {
        return ExtendedSocketOptions.PlatformSocketOptions.get().flowSupported();
    }
    
    private static boolean keepAliveOptionsSupported() {
        return ExtendedSocketOptions.PlatformSocketOptions.get().keepAliveOptionsSupported();
    }
    
    private static void setTcpkeepAliveProbes(FileDescriptor fileDescriptor, int n) throws SocketException {
        ExtendedSocketOptions.PlatformSocketOptions.get().setTcpkeepAliveProbes(ExtendedSocketOptions.fdAccess.get(fileDescriptor), n);
    }
    
    private static void setTcpKeepAliveTime(FileDescriptor fileDescriptor, int n) throws SocketException {
        ExtendedSocketOptions.PlatformSocketOptions.get().setTcpKeepAliveTime(ExtendedSocketOptions.fdAccess.get(fileDescriptor), n);
    }
    
    private static void setTcpKeepAliveIntvl(FileDescriptor fileDescriptor, int n) throws SocketException {
        ExtendedSocketOptions.PlatformSocketOptions.get().setTcpKeepAliveIntvl(ExtendedSocketOptions.fdAccess.get(fileDescriptor), n);
    }
    
    private static int getTcpkeepAliveProbes(FileDescriptor fileDescriptor) throws SocketException {
        return ExtendedSocketOptions.PlatformSocketOptions.get().getTcpkeepAliveProbes(ExtendedSocketOptions.fdAccess.get(fileDescriptor));
    }
    
    private static int getTcpKeepAliveTime(FileDescriptor fileDescriptor) throws SocketException {
        return ExtendedSocketOptions.PlatformSocketOptions.get().getTcpKeepAliveTime(ExtendedSocketOptions.fdAccess.get(fileDescriptor));
    }
    
    private static int getTcpKeepAliveIntvl(FileDescriptor fileDescriptor) throws SocketException {
        return ExtendedSocketOptions.PlatformSocketOptions.get().getTcpKeepAliveIntvl(ExtendedSocketOptions.fdAccess.get(fileDescriptor));
    }
    
    static Object access$100(Object o, Class clazz) {
        return checkValueType(o, clazz);
    }
    
    static void access$200(FileDescriptor fileDescriptor, int n) throws SocketException {
        setTcpkeepAliveProbes(fileDescriptor, n);
    }
    
    static void access$300(FileDescriptor fileDescriptor, int n) throws SocketException {
        setTcpKeepAliveTime(fileDescriptor, n);
    }
    
    static void access$400(FileDescriptor fileDescriptor, int n) throws SocketException {
        setTcpKeepAliveIntvl(fileDescriptor, n);
    }
    
    static int access$500(FileDescriptor fileDescriptor) throws SocketException {
        return getTcpkeepAliveProbes(fileDescriptor);
    }
    
    static int access$600(FileDescriptor fileDescriptor) throws SocketException {
        return getTcpKeepAliveTime(fileDescriptor);
    }
    
    static int access$700(FileDescriptor fileDescriptor) throws SocketException {
        return getTcpKeepAliveIntvl(fileDescriptor);
    }
    
    static {
        SO_FLOW_SLA = new ExtendedSocketOptions.ExtSocketOption("SO_FLOW_SLA", SocketFlow.class);
        TCP_KEEPIDLE = new ExtendedSocketOptions.ExtSocketOption("TCP_KEEPIDLE", Integer.class);
        TCP_KEEPINTERVAL = new ExtendedSocketOptions.ExtSocketOption("TCP_KEEPINTERVAL", Integer.class);
        TCP_KEEPCOUNT = new ExtendedSocketOptions.ExtSocketOption("TCP_KEEPCOUNT", Integer.class);
        platformSocketOptions = ExtendedSocketOptions.PlatformSocketOptions.get();
        flowSupported = ExtendedSocketOptions.platformSocketOptions.flowSupported();
        keepAliveOptSupported = ExtendedSocketOptions.platformSocketOptions.keepAliveOptionsSupported();
        extendedOptions = options();
        sun.net.ExtendedSocketOptions.register(new sun.net.ExtendedSocketOptions(ExtendedSocketOptions.extendedOptions) {
            public void setOption(FileDescriptor fileDescriptor, SocketOption<?> obj, Object o) throws SocketException {
                SecurityManager securityManager = System.getSecurityManager();
                if (securityManager != null) {
                    securityManager.checkPermission(new NetworkPermission("setOption." + obj.name()));
                }
                if (fileDescriptor == null || !fileDescriptor.valid()) {
                    throw new SocketException("socket closed");
                }
                if (obj == ExtendedSocketOptions.SO_FLOW_SLA) {
                    assert !(!ExtendedSocketOptions.flowSupported);
                    ExtendedSocketOptions.setFlowOption(fileDescriptor, (SocketFlow)ExtendedSocketOptions.access$100(o, SocketFlow.class));
                }
                else if (obj == ExtendedSocketOptions.TCP_KEEPCOUNT) {
                    ExtendedSocketOptions.access$200(fileDescriptor, (int)o);
                }
                else if (obj == ExtendedSocketOptions.TCP_KEEPIDLE) {
                    ExtendedSocketOptions.access$300(fileDescriptor, (int)o);
                }
                else if (obj == ExtendedSocketOptions.TCP_KEEPINTERVAL) {
                    ExtendedSocketOptions.access$400(fileDescriptor, (int)o);
                }
                else {
                    throw new InternalError("Unexpected option " + obj);
                }
                return;
            }
            
            public Object getOption(FileDescriptor fileDescriptor, SocketOption<?> obj) throws SocketException {
                SecurityManager securityManager = System.getSecurityManager();
                if (securityManager != null) {
                    securityManager.checkPermission(new NetworkPermission("getOption." + obj.name()));
                }
                if (fileDescriptor == null || !fileDescriptor.valid()) {
                    throw new SocketException("socket closed");
                }
                if (obj == ExtendedSocketOptions.SO_FLOW_SLA) {
                    assert !(!ExtendedSocketOptions.flowSupported);
                    SocketFlow create = SocketFlow.create();
                    ExtendedSocketOptions.getFlowOption(fileDescriptor, create);
                    return create;
                }
                else {
                    if (obj == ExtendedSocketOptions.TCP_KEEPCOUNT) {
                        return Integer.valueOf(ExtendedSocketOptions.access$500(fileDescriptor));
                    }
                    if (obj == ExtendedSocketOptions.TCP_KEEPIDLE) {
                        return Integer.valueOf(ExtendedSocketOptions.access$600(fileDescriptor));
                    }
                    if (obj == ExtendedSocketOptions.TCP_KEEPINTERVAL) {
                        return Integer.valueOf(ExtendedSocketOptions.access$700(fileDescriptor));
                    }
                    throw new InternalError("Unexpected option " + obj);
                }
            }
        });
        fdAccess = SharedSecrets.getJavaIOFileDescriptorAccess();
    }
    
    private static class ExtSocketOption<T> implements SocketOption<T>
    {
        private final String name;
        private final Class<T> type;
        
        ExtSocketOption(String name, Class<T> type) {
            this.name = name;
            this.type = type;
        }
        
        @Override
        public String name() {
            return this.name;
        }
        
        @Override
        public Class<T> type() {
            return this.type;
        }
        
        @Override
        public String toString() {
            return this.name;
        }
    }
    
    static class PlatformSocketOptions
    {
        private static final ExtendedSocketOptions.PlatformSocketOptions instance;
        
        protected PlatformSocketOptions() {
        }
        
        private static ExtendedSocketOptions.PlatformSocketOptions newInstance(String className) {
            try {
                return (ExtendedSocketOptions.PlatformSocketOptions)Class.forName(className).getConstructor(new Class[0]).newInstance(new Object[0]);
            }
            catch (ReflectiveOperationException detailMessage) {
                throw new AssertionError(detailMessage);
            }
        }
        
        private static ExtendedSocketOptions.PlatformSocketOptions create() {
            String s = (String)AccessController.doPrivileged(new PrivilegedAction<String>() {
                @Override
                public String run() {
                    return System.getProperty("os.name");
                }
            });
            if ("SunOS".equals(s)) {
                return newInstance("jdk.net.SolarisSocketOptions");
            }
            if ("Linux".equals(s)) {
                return newInstance("jdk.net.LinuxSocketOptions");
            }
            if (s.startsWith("Mac")) {
                return newInstance("jdk.net.MacOSXSocketOptions");
            }
            return new ExtendedSocketOptions.PlatformSocketOptions();
        }
        
        static ExtendedSocketOptions.PlatformSocketOptions get() {
            return ExtendedSocketOptions.PlatformSocketOptions.instance;
        }
        
        void init() {
        }
        
        int setFlowOption(int n, int n2, long n3) throws SocketException {
            throw new UnsupportedOperationException("unsupported socket option");
        }
        
        int getFlowOption(int n, SocketFlow socketFlow) throws SocketException {
            throw new UnsupportedOperationException("unsupported socket option");
        }
        
        boolean flowSupported() {
            return false;
        }
        
        boolean keepAliveOptionsSupported() {
            return false;
        }
        
        void setTcpkeepAliveProbes(int n, int n2) throws SocketException {
            throw new UnsupportedOperationException("unsupported TCP_KEEPCNT option");
        }
        
        void setTcpKeepAliveTime(int n, int n2) throws SocketException {
            throw new UnsupportedOperationException("unsupported TCP_KEEPIDLE option");
        }
        
        void setTcpKeepAliveIntvl(int n, int n2) throws SocketException {
            throw new UnsupportedOperationException("unsupported TCP_KEEPINTVL option");
        }
        
        int getTcpkeepAliveProbes(int n) throws SocketException {
            throw new UnsupportedOperationException("unsupported TCP_KEEPCNT option");
        }
        
        int getTcpKeepAliveTime(int n) throws SocketException {
            throw new UnsupportedOperationException("unsupported TCP_KEEPIDLE option");
        }
        
        int getTcpKeepAliveIntvl(int n) throws SocketException {
            throw new UnsupportedOperationException("unsupported TCP_KEEPINTVL option");
        }
        
        static {
            instance = create();
        }
    }
}
